from typing import Any, List, Optional, Sequence, Set, Tuple, Type
from uuid import UUID

import pytest
from guardpost.authentication import Identity
from pytest import raises
from rodi import Container

from blacksheep import FormContent, FormPart, JSONContent, MultiPartFormData, Request
from blacksheep.server.bindings import (
    BadRequest,
    Binder,
    BinderAlreadyDefinedException,
    BinderNotRegisteredForValueType,
    BodyBinder,
    BoundValue,
    CookieBinder,
    FormBinder,
    HeaderBinder,
    IdentityBinder,
    InvalidRequestBody,
    JSONBinder,
    MissingBodyError,
    MissingConverterError,
    NameAliasAlreadyDefinedException,
    QueryBinder,
    RequestMethodBinder,
    RequestURLBinder,
    RouteBinder,
    ServiceBinder,
    SyncBinder,
    TypeAliasAlreadyDefinedException,
    get_binder_by_type,
)
from blacksheep.url import URL

JSONContentType = (b"Content-Type", b"application/json")


class ExampleOne:
    def __init__(self, a, b):
        self.a = a
        self.b = int(b)


class ExampleTwo:
    def __init__(self, a, b, **kwargs):
        self.a = a
        self.b = b


class ExampleThree:
    def __init__(self, a: str, b: List[str]):
        self.a = a
        self.b = b


@pytest.mark.asyncio
async def test_from_body_json_binding():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent({"a": "world", "b": 9000})
    )

    parameter = JSONBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


@pytest.mark.asyncio
async def test_from_body_json_binding_extra_parameters_strategy():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            {
                "a": "world",
                "b": 9000,
                "c": "This is an extra parameter, accepted by constructor explicitly",
            }
        )
    )

    parameter = JSONBinder(ExampleTwo)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleTwo)
    assert value.a == "world"
    assert value.b == 9000


@pytest.mark.asyncio
async def test_from_body_json_with_converter():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent(
            {
                "a": "world",
                "b": 9000,
                "c": "This is an extra parameter, accepted by constructor explicitly",
            }
        )
    )

    def convert(data):
        return ExampleOne(data.get("a"), data.get("b"))

    parameter = JSONBinder(ExampleOne, converter=convert)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


@pytest.mark.asyncio
async def test_from_body_json_binding_request_missing_content_type():
    request = Request("POST", b"/", [])

    parameter = JSONBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert value is None


@pytest.mark.asyncio
async def test_from_body_json_binding_invalid_input():
    request = Request("POST", b"/", [JSONContentType]).with_content(
        JSONContent({"c": 1, "d": 2})
    )

    parameter = JSONBinder(ExampleOne)

    with raises(BadRequest):
        await parameter.get_value(request)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,header_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"\xc5\x81ukasz", "Łukasz"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
    ],
)
async def test_from_header_binding(expected_type, header_value, expected_value):
    request = Request("GET", b"/", [(b"X-Foo", header_value)])

    parameter = HeaderBinder(expected_type, "X-Foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,header_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"\xc5\x81ukasz", "Łukasz"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
    ],
)
async def test_from_header_binding_name_ci(expected_type, header_value, expected_value):
    request = Request("GET", b"/", [(b"X-Foo", header_value)])

    parameter = HeaderBinder(expected_type, "x-foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,query_value,expected_value",
    [
        [str, b"Foo", "Foo"],
        [str, b"foo", "foo"],
        [str, b"Hello%20World%21%3F", "Hello World!?"],
        [int, b"1", 1],
        [int, b"10", 10],
        [float, b"1.5", 1.5],
        [float, b"1241.5", 1241.5],
        [bool, b"1", True],
        [bool, b"0", False],
    ],
)
async def test_from_query_binding(expected_type, query_value, expected_value):
    request = Request("GET", b"/?foo=" + query_value, None)

    parameter = QueryBinder(expected_type, "foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_value


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,route_value,expected_value",
    [
        [str, "Foo", "Foo"],
        [str, "foo", "foo"],
        [str, "Hello%20World%21%3F", "Hello World!?"],
        [int, "1", 1],
        [int, "10", 10],
        [float, "1.5", 1.5],
        [float, "1241.5", 1241.5],
        [bool, "1", True],
        [bool, "0", False],
        [
            UUID,
            "b0c1f822-b63c-475e-9f2e-b6406bafcc2b",
            UUID("b0c1f822-b63c-475e-9f2e-b6406bafcc2b"),
        ],
    ],
)
async def test_from_route_binding(expected_type, route_value, expected_value):
    request = Request("GET", b"/", None)
    request.route_values = {"name": route_value}

    parameter = RouteBinder(expected_type, "name")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_value


@pytest.mark.asyncio
@pytest.mark.parametrize("binder_type", [HeaderBinder, QueryBinder, RouteBinder])
async def test_raises_for_missing_default_converter(binder_type):
    with raises(MissingConverterError):
        binder_type("example", ExampleOne)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,invalid_value",
    [[int, "x"], [int, ""], [float, "x"], [float, ""], [bool, "x"], [bool, "yes"]],
)
async def test_from_route_raises_for_invalid_parameter(expected_type, invalid_value):
    request = Request("GET", b"/", None)
    request.route_values = {"name": invalid_value}

    parameter = RouteBinder(expected_type, "name")

    with raises(BadRequest):
        await parameter.get_value(request)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_type,invalid_value",
    [
        [int, b"x"],
        [int, b""],
        [float, b"x"],
        [float, b""],
        [bool, b"x"],
        [bool, b"yes"],
    ],
)
async def test_from_query_raises_for_invalid_parameter(
    expected_type, invalid_value: bytes
):
    request = Request("GET", b"/?foo=" + invalid_value, None)

    parameter = QueryBinder(expected_type, "foo", required=True)

    with raises(BadRequest):
        await parameter.get_value(request)


@pytest.mark.asyncio
async def test_from_services():
    request = Request("GET", b"/", [])

    service_instance = ExampleOne(1, 2)
    container = Container()
    container.add_instance(service_instance)

    parameter = ServiceBinder(ExampleOne, "service", False, container.build_provider())
    value = await parameter.get_value(request)

    assert value is service_instance


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "declared_type,expected_type,header_values,expected_values",
    [
        [List[str], list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [
            Tuple[str],
            tuple,
            [b"Lorem", b"ipsum", b"dolor"],
            ("Lorem", "ipsum", "dolor"),
        ],
        [Set[str], set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [
            Sequence[str],
            list,
            [b"Lorem", b"ipsum", b"dolor"],
            ["Lorem", "ipsum", "dolor"],
        ],
    ],
)
async def test_from_header_binding_iterables(
    declared_type, expected_type, header_values, expected_values
):
    request = Request("GET", b"/", [(b"X-Foo", value) for value in header_values])

    parameter = HeaderBinder(declared_type, "X-Foo")

    value = await parameter.get_value(request)

    assert isinstance(value, expected_type)
    assert value == expected_values


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "declared_type,expected_type,query_values,expected_values",
    [
        [list, list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [tuple, tuple, [b"Lorem", b"ipsum", b"dolor"], ("Lorem", "ipsum", "dolor")],
        [set, set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [List, list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [Tuple, tuple, [b"Lorem", b"ipsum", b"dolor"], ("Lorem", "ipsum", "dolor")],
        [Set, set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [List[str], list, [b"Lorem", b"ipsum", b"dolor"], ["Lorem", "ipsum", "dolor"]],
        [
            Tuple[str],
            tuple,
            [b"Lorem", b"ipsum", b"dolor"],
            ("Lorem", "ipsum", "dolor"),
        ],
        [Set[str], set, [b"Lorem", b"ipsum", b"dolor"], {"Lorem", "ipsum", "dolor"}],
        [
            Sequence[str],
            list,
            [b"Lorem", b"ipsum", b"dolor"],
            ["Lorem", "ipsum", "dolor"],
        ],
        [List[int], list, [b"10"], [10]],
        [List[int], list, [b"0", b"1", b"0"], [0, 1, 0]],
        [List[int], list, [b"0", b"1", b"0", b"2"], [0, 1, 0, 2]],
        [List[bytes], list, [b"0", b"1", b"0", b"2"], [b"0", b"1", b"0", b"2"]],
        [List[bool], list, [b"1"], [True]],
        [List[bool], list, [b"0", b"1", b"0"], [False, True, False]],
        [List[bool], list, [b"0", b"1", b"0", b"true"], [False, True, False, True]],
        [List[float], list, [b"10.2"], [10.2]],
        [List[float], list, [b"0.3", b"1", b"0"], [0.3, 1.0, 0]],
        [List[float], list, [b"0.5", b"1", b"0", b"2"], [0.5, 1.0, 0, 2.0]],
        [Tuple[float], tuple, [b"10.2"], (10.2,)],
        [Tuple[float], tuple, [b"0.3", b"1", b"0"], (0.3, 1.0, 0)],
        [Tuple[float], tuple, [b"0.5", b"1", b"0", b"2"], (0.5, 1.0, 0, 2.0)],
        [Set[int], set, [b"10"], {10}],
        [Set[int], set, [b"0", b"1", b"0"], {0, 1, 0}],
        [Set[int], set, [b"0", b"1", b"0", b"2"], {0, 1, 0, 2}],
        [
            List[UUID],
            list,
            [
                b"de18d268-f5c5-42db-89b2-c61bbfe96e65",
                b"d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe",
                b"00000000-0000-0000-0000-000000000000",
            ],
            [
                UUID("de18d268-f5c5-42db-89b2-c61bbfe96e65"),
                UUID("d5fd0cde-4ad6-4b61-a5b1-5b8e6d48cebe"),
                UUID("00000000-0000-0000-0000-000000000000"),
            ],
        ],
    ],
)
async def test_from_query_binding_iterables(
    declared_type, expected_type, query_values, expected_values
):
    qs = b"&foo=".join([value for value in query_values])

    request = Request("GET", b"/?foo=" + qs, None)

    parameter = QueryBinder(declared_type, "foo")

    values = await parameter.get_value(request)

    assert isinstance(values, expected_type)
    assert values == expected_values


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "declared_type", [List[List[str]], Tuple[Tuple[str]], List[list]]
)
async def test_nested_iterables_raise_missing_converter_from_header(declared_type):
    with raises(MissingConverterError):
        HeaderBinder(declared_type)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "declared_type", [List[List[str]], Tuple[Tuple[str]], List[list]]
)
async def test_nested_iterables_raise_missing_converter_from_query(declared_type):
    with raises(MissingConverterError):
        QueryBinder("example", declared_type)


@pytest.mark.asyncio
async def test_identity_binder_identity_not_set():
    request = Request("GET", b"/", None)

    parameter = IdentityBinder(Identity)

    value = await parameter.get_value(request)

    assert value is None


@pytest.mark.asyncio
async def test_identity_binder():
    request = Request("GET", b"/", None)
    request.identity = Identity({})

    parameter = IdentityBinder(Identity)

    value = await parameter.get_value(request)

    assert value is request.identity


@pytest.mark.asyncio
async def test_from_body_form_binding_urlencoded():
    request = Request("POST", b"/", []).with_content(
        FormContent({"a": "world", "b": 9000})
    )

    parameter = FormBinder(ExampleOne)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


@pytest.mark.asyncio
async def test_from_body_form_binding_urlencoded_keys_duplicates():
    request = Request("POST", b"/", []).with_content(
        FormContent([("a", "world"), ("b", "one"), ("b", "two"), ("b", "three")])
    )

    parameter = FormBinder(ExampleThree)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleThree)
    assert value.a == "world"
    assert value.b == ["one", "two", "three"]


@pytest.mark.asyncio
async def test_from_body_form_binding_multipart():
    request = Request("POST", b"/", []).with_content(
        MultiPartFormData([FormPart(b"a", b"world"), FormPart(b"b", b"9000")])
    )

    parameter = FormBinder(ExampleOne)
    value = await parameter.get_value(request)

    assert isinstance(value, ExampleOne)
    assert value.a == "world"
    assert value.b == 9000


@pytest.mark.asyncio
async def test_from_body_form_binding_multipart_keys_duplicates():
    request = Request("POST", b"/", []).with_content(
        MultiPartFormData(
            [
                FormPart(b"a", b"world"),
                FormPart(b"b", b"one"),
                FormPart(b"b", b"two"),
                FormPart(b"b", b"three"),
            ]
        )
    )

    parameter = FormBinder(ExampleThree)

    value = await parameter.get_value(request)

    assert isinstance(value, ExampleThree)
    assert value.a == "world"
    assert value.b == ["one", "two", "three"]


@pytest.mark.asyncio
async def test_custom_bound_value_and_binder():
    class FromMethod(BoundValue[str]):
        pass

    class MethodBinder(Binder):
        handle = FromMethod

        async def get_value(self, request: Request) -> Optional[str]:
            return request.method

    parameter = MethodBinder(str)

    for method in {"GET", "POST", "TRACE"}:
        value = await parameter.get_value(Request(method, b"/", []))
        assert value == method


@pytest.mark.asyncio
async def test_custom_bound_value_fails_for_missing_binder():
    class FromSomething(BoundValue[str]):
        pass

    def faulty(example: FromSomething):
        pass

    with pytest.raises(BinderNotRegisteredForValueType):
        get_binder_by_type(FromSomething)


@pytest.mark.asyncio
async def test_raises_for_duplicate_binders():
    class FromMethod(BoundValue[str]):
        pass

    class MethodBinder(Binder):
        handle = FromMethod

        async def get_value(self, request: Request) -> Optional[str]:
            return request.method

    with pytest.raises(BinderAlreadyDefinedException):

        class MethodBinder2(Binder):
            handle = FromMethod

            async def get_value(self, request: Request) -> Optional[str]:
                return request.method


@pytest.mark.parametrize(
    "binder_type,expected_source_name",
    [(RouteBinder, "route"), (QueryBinder, "query"), (HeaderBinder, "header")],
)
def test_sync_binder_source_name(
    binder_type: Type[SyncBinder], expected_source_name: str
):
    binder = binder_type(str)
    assert binder.source_name == expected_source_name


@pytest.mark.asyncio
async def test_body_binder_throws_for_abstract_methods():
    body_binder = BodyBinder(dict)

    with pytest.raises(NotImplementedError):
        body_binder.matches_content_type(Request("HEAD", b"/", []))

    with pytest.raises(NotImplementedError):
        await body_binder.read_data(Request("HEAD", b"/", []))


@pytest.mark.asyncio
async def test_body_binder_throws_bad_request_for_missing_body():
    class CustomBodyBinder(BodyBinder):
        def matches_content_type(self, request: Request) -> bool:
            return True

        async def read_data(self, request: Request) -> Any:
            return None

    body_binder = CustomBodyBinder(dict)

    with pytest.raises(MissingBodyError):
        await body_binder.get_value(
            Request("POST", b"/", [(b"content-type", b"application/json")])
        )

    body_binder = JSONBinder(dict, required=True)

    with pytest.raises(MissingBodyError):
        await body_binder.get_value(Request("POST", b"/", []))


@pytest.mark.asyncio
async def test_body_binder_throws_bad_request_for_value_error():
    body_binder = JSONBinder(dict, required=True)

    def example_converter(value):
        raise ValueError("Invalid value")

    body_binder.converter = example_converter

    with pytest.raises(InvalidRequestBody):
        await body_binder.get_value(
            Request(
                "POST", b"/", [(b"content-type", b"application/json")]
            ).with_content(JSONContent({"id": "1", "name": "foo"}))
        )


def test_sync_binders_source_name():
    assert CookieBinder().source_name == "cookie"
    assert HeaderBinder().source_name == "header"
    assert QueryBinder().source_name == "query"
    assert RouteBinder().source_name == "route"


@pytest.mark.parametrize("method", ["GET", "OPTIONS", "POST"])
@pytest.mark.asyncio
async def test_request_method_binder(method):
    request = Request(method, b"/", [])
    parameter = RequestMethodBinder()
    value = await parameter.get_value(request)
    assert value == method


@pytest.mark.parametrize("url", [b"/", b"/api/cats/123", b"/foo/index.html?s=20"])
@pytest.mark.asyncio
async def test_request_url_binder(url):
    request = Request("GET", url, [])
    parameter = RequestURLBinder()
    value = await parameter.get_value(request)
    assert value == URL(url)


def test_duplicate_name_alias_raises():
    class FooBinder1(Binder):
        name_alias = "foo_absurd_example"

    with pytest.raises(NameAliasAlreadyDefinedException) as duplicate_alias_error:

        class FooBinder2(Binder):
            name_alias = "foo_absurd_example"

    assert str(duplicate_alias_error.value) == (
        "There is already a name alias defined for 'foo_absurd_example', "
        "the second type is: FooBinder2"
    )


def test_duplicate_type_alias_raises():
    class X:
        pass

    class XBinder1(Binder):
        type_alias = X

    with pytest.raises(TypeAliasAlreadyDefinedException) as duplicate_alias_error:

        class XBinder2(Binder):
            type_alias = X

    assert str(duplicate_alias_error.value) == (
        "There is already a type alias defined for 'X', the second type is: XBinder2"
    )



from collections.abc import Mapping, MutableSequence
from typing import Dict, List, Tuple, Union


cdef class Header:

    def __init__(self, bytes name, bytes value):
        self.name = name
        self.value = value

    def __repr__(self):
        return f'<Header {self.name}: {self.value}>'

    def __iter__(self):
        yield self.name
        yield self.value

    def __eq__(self, other):
        if isinstance(other, Header):
            return other.name.lower() == self.name.lower() and other.value == self.value
        return NotImplemented


cdef class Headers:

    def __init__(self, list values = None):
        if values is None:
            values = []
        self.values = values

    cpdef tuple get(self, bytes name):
        cdef list results = []
        cdef tuple header
        name = name.lower()
        for header in self.values:
            if header[0].lower() == name:
                results.append(header[1])
        return tuple(results)

    cpdef list get_tuples(self, bytes name):
        cdef list results = []
        cdef tuple header
        name = name.lower()
        for header in self.values:
            if header[0].lower() == name:
                results.append(header)
        return results

    cpdef bytes get_first(self, bytes key):
        cdef tuple header
        key = key.lower()
        for header in self.values:
            if header[0].lower() == key:
                return header[1]

    cpdef bytes get_single(self, bytes key):
        cdef tuple results = self.get(key)
        if len(results) > 1:
            raise ValueError('Headers contains more than one header with the given key')
        if len(results) < 1:
            raise ValueError('Headers does not contain one header with the given key')
        return results[0]

    cpdef void merge(self, list values):
        cdef tuple header
        for header in values:
            if header is None:
                continue
            self.values.append(header)

    def update(self, dict values: Dict[bytes, bytes]):
        for key, value in values.items():
            self[key] = value

    def items(self):
        yield from self.values

    cpdef Headers clone(self):
        cdef list values = []
        cdef bytes name, value
        for name, value in self.values:
            values.append((name, value))
        return Headers(values)

    def add_many(self, values: Union[Dict[bytes, bytes], List[Tuple[bytes, bytes]]]):
        if isinstance(values, MutableSequence):
            for item in values:
                self.add(*item)
            return

        if isinstance(values, Mapping):
            for key, value in values.items():
                self.add(key, value)
            return
        raise ValueError('values must be Dict[bytes, bytes] or List[Header]')

    @staticmethod
    def _add_to_instance(instance, other):
        if isinstance(other, Headers):
            for value in other:
                instance.add(*value)
            return instance

        if isinstance(other, Header):
            instance.add(other.name, other.value)
            return instance

        if isinstance(other, tuple):
            if len(other) != 2:
                raise ValueError(f'Cannot add, an invalid tuple {str(other)}.')
            instance.add(*other)
            return instance

        if isinstance(other, MutableSequence):
            for value in other:
                if isinstance(value, tuple) and len(value) == 2:
                    instance.add(*value)
                else:
                    raise ValueError(f'The sequence contains invalid elements: '
                                     f'cannot add {str(value)} to {instance.__class__.__name__}')
            return instance

        return NotImplemented

    def __add__(self, other):
        return self._add_to_instance(self.clone(), other)

    def __radd__(self, other):
        return self._add_to_instance(self.clone(), other)

    def __iadd__(self, other):
        return self._add_to_instance(self, other)

    def __iter__(self):
        yield from self.values

    def __setitem__(self, bytes key, bytes value):
        # Not obvious, but here we make the decision that setter removes existing headers with matching name:
        # it feels more natural with syntax: headers[b'X-Foo'] = b'Something'
        self.set(key, value)

    def __getitem__(self, bytes item):
        return self.get(item)

    cpdef tuple keys(self):
        cdef bytes name, value
        cdef list results = []

        for name, value in self.values:
            if name not in results:
                results.append(name)
        return tuple(results)

    cpdef void add(self, bytes name, bytes value):
        self.values.append((name, value))

    cpdef void set(self, bytes name, bytes value):
        if self.contains(name):
            self.remove(name)
        self.add(name, value)

    cpdef void remove(self, bytes key):
        cdef tuple item
        cdef list to_remove = []
        key = key.lower()

        for item in self.values:
            if item[0].lower() == key:
                to_remove.append(item)

        for item in to_remove:
            self.values.remove(item)

    cpdef bint contains(self, bytes key):
        cdef bytes name, value
        key = key.lower()

        for name, value in self.values:
            if name.lower() == key:
                return True
        return False

    def __delitem__(self, bytes key):
        self.remove(key)

    def __contains__(self, bytes key):
        return self.contains(key)

    def __repr__(self):
        return f'<Headers {self.values}>'



import pytest

from blacksheep.server.application import Application
from blacksheep.server.routing import (
    InvalidValuePatternName,
    MountRegistry,
    Route,
    RouteDuplicate,
    RouteException,
    RouteMethod,
    Router,
)

FAKE = b"FAKE"

MATCHING_ROUTES = [
    ("head", b"", b"/"),
    ("get", b"", b"/"),
    ("head", b"/", b"/"),
    ("get", b"/", b"/"),
    ("get", b"/:a", b"/foo"),
    ("get", b"/foo", b"/foo"),
    ("get", b"/foo", b"/Foo"),
    ("get", b"/:a/:b", b"/foo/oof"),
    ("post", b"/", b"/"),
    ("post", b"/:id", b"/123"),
    ("put", b"/", b"/"),
    ("delete", b"/", b"/"),
]

NON_MATCHING_ROUTE = [
    ("head", b"/", b"/foo"),
    ("get", b"/", b"/foo"),
    ("post", b"/", b"/foo"),
    ("post", b"/foo", b"/123"),
    ("put", b"/a/b/c/d", b"/a/b/c/"),
    ("put", b"/a/b/c/d", b"/a/b/c/d/e"),
    ("delete", b"/", b"/a"),
]


def mock_handler():
    return None


class MockHandler:
    def __init__(self, request_handler, auth_handler):
        self.request_handler = request_handler
        self.auth_handler = auth_handler


@pytest.mark.parametrize(
    "pattern,url,expected_values",
    [
        (b"/foo/:id", b"/foo/123", {"id": "123"}),
        (b"/foo/{id}", b"/foo/123", {"id": "123"}),
        (b"/foo/<id>", b"/foo/123", {"id": "123"}),
        ("/foo/:id", b"/foo/123", {"id": "123"}),
        ("/foo/{id}", b"/foo/123", {"id": "123"}),
        ("/foo/<id>", b"/foo/123", {"id": "123"}),
        (b"/foo/:id/ufo/:b", b"/foo/223/ufo/a13", {"id": "223", "b": "a13"}),
        (b"/foo/{id}/ufo/{b}", b"/foo/223/ufo/a13", {"id": "223", "b": "a13"}),
        (b"/foo/<id>/ufo/<b>", b"/foo/223/ufo/a13", {"id": "223", "b": "a13"}),
        ("/foo/:id/ufo/:b", b"/foo/223/ufo/a13", {"id": "223", "b": "a13"}),
        ("/foo/{id}/ufo/{b}", b"/foo/223/ufo/a13", {"id": "223", "b": "a13"}),
        (b"/foo/:id/ufo/:b", b"/Foo/223/Ufo/a13", {"id": "223", "b": "a13"}),
        (b"/:a", b"/Something", {"a": "Something"}),
        (b"/{a}", b"/Something", {"a": "Something"}),
        (b"/<a>", b"/Something", {"a": "Something"}),
        (b"/alive", b"/alive", None),
    ],
)
def test_route_good_matches(pattern, url, expected_values):
    route = Route(pattern, mock_handler)
    print(route.full_pattern)
    match = route.match(url)

    assert match is not None
    assert match.values == expected_values


@pytest.mark.parametrize(
    "pattern,url,expected_values",
    [
        ("/foo/{string:foo}", b"/foo/Hello", {"foo": "Hello"}),
        ("/foo/{string:foo}/ufo", b"/foo/Hello/ufo", {"foo": "Hello"}),
        ("/foo/{str:foo}", b"/foo/Hello", {"foo": "Hello"}),
        ("/foo/{str:foo}/ufo", b"/foo/Hello/ufo", {"foo": "Hello"}),
        ("/foo/{int:id}", b"/foo/123", {"id": "123"}),
        ("/foo/<int:id>", b"/foo/123", {"id": "123"}),
        ("/foo/{float:a}", b"/foo/123", {"a": "123"}),
        ("/foo/{float:a}", b"/foo/123.15", {"a": "123.15"}),
        ("/foo/<int:a>/<float:b>", b"/foo/123/777.77", {"a": "123", "b": "777.77"}),
        (
            "/foo/{uuid:id}",
            b"/foo/52464abf-f583-4b32-80f8-704bcb9e36a2",
            {"id": "52464abf-f583-4b32-80f8-704bcb9e36a2"},
        ),
        (
            "/public/{path:file}",
            b"/public/",
            {"file": ""},
        ),
        (
            "/public/{path:file}",
            b"/public/js/home/home.js",
            {"file": "js/home/home.js"},
        ),
        (
            "/public/{path:file}",
            b"/public/home.js",
            {"file": "home.js"},
        ),
        (
            "/public/<path:file>",
            b"/public/js/home/home.js",
            {"file": "js/home/home.js"},
        ),
        (
            "/<path:filepath>",
            b"/public/js/home/home.js",
            {"filepath": "public/js/home/home.js"},
        ),
    ],
)
def test_route_good_matches_with_parameter_patterns(pattern, url, expected_values):
    route = Route(pattern, mock_handler)
    match = route.match(url)

    assert match is not None
    assert match.values == expected_values


@pytest.mark.parametrize(
    "pattern,url",
    [
        (b"/foo/{int:id}", b"/foo/abc"),
        (b"/foo/<int:id>", b"/foo/X"),
        (b"/foo/{float:a}", b"/foo/false"),
        (b"/foo/{float:a}", b"/foo/123.15.31"),
        (b"/foo/{float:a}", b"/foo/123,50"),
        (b"/foo/{uuid:a}", b"/foo/not-a-guid"),
    ],
)
def test_route_bad_matches_with_parameter_patterns(pattern, url):
    route = Route(pattern, mock_handler)
    match = route.match(url)
    assert match is None


@pytest.mark.parametrize(
    "pattern",
    (
        "/public/<path:file>",
        "/<path:file>",
        "/public/*",
        "/*",
        "*",
    ),
)
def test_sort_routes_path_pattern(pattern):
    router = Router()
    catch_all_route = Route(pattern, mock_handler)

    router.add_route("GET", catch_all_route)
    router.add_route("GET", Route("/cat/:cat_id", mock_handler))
    router.add_route("GET", Route("/index", mock_handler))
    router.add_route("GET", Route("/about", mock_handler))

    assert router.routes[b"GET"][0] is catch_all_route

    router.sort_routes()

    assert router.routes[b"GET"][-1] is catch_all_route


@pytest.mark.parametrize(
    "pattern,invalid_pattern_name",
    [
        (b"/foo/{xxx:id}", "xxx"),
        (b"/foo/{ind:id}", "ind"),
    ],
)
def test_route_raises_for_invalid_parameter_name(pattern, invalid_pattern_name):
    with pytest.raises(InvalidValuePatternName):
        Route(pattern, mock_handler)


@pytest.mark.parametrize(
    "pattern",
    ["/a", "/a/b", "/a/b/c", "/cats/:cat_id", "/cats/:cat_id/friends"],
)
def test_route_repr(pattern: str):
    route = Route(pattern, mock_handler)
    assert repr(route) == f"<Route {pattern}>"


@pytest.mark.parametrize(
    "pattern,url",
    [
        (b"/foo/:id", b"/fo/123"),
        (b"/foo/:id/ufo/:b", b"/foo/223/uof/a13"),
        (b"/:a", b"/"),
    ],
)
def test_route_bad_matches(pattern, url):
    route = Route(pattern, mock_handler)
    match = route.match(url)

    assert match is None


@pytest.mark.parametrize("pattern", [b"/:a/:a", b"/foo/:a/ufo/:a", b"/:foo/a/:foo"])
def test_invalid_route_repeated_group_name(pattern):
    with pytest.raises(ValueError):
        Route(pattern, mock_handler)


def test_route_handler_can_be_anything():
    def request_handler():
        pass

    def auth_handler():
        pass

    handler = MockHandler(request_handler, auth_handler)

    route = Route(b"/", handler)
    match = route.match(b"/")

    assert match is not None
    assert match.handler.request_handler is request_handler
    assert match.handler.auth_handler is auth_handler


@pytest.mark.parametrize("method,pattern,url", MATCHING_ROUTES)
def test_router_add_method(method, pattern, url):
    router = Router()
    router.add(method, pattern, mock_handler)

    match = router.get_match(method, url)

    assert match is not None
    assert match.handler is mock_handler

    route = router.get_matching_route(method, url)
    assert route is not None

    match = router.get_match(FAKE, url)
    assert match is None

    route = router.get_matching_route(FAKE, url)
    assert route is None


@pytest.mark.parametrize("method,pattern,url", NON_MATCHING_ROUTE)
def test_router_not_matching_routes(method, pattern, url):
    router = Router()
    router.add(method, pattern, mock_handler)
    route = router.get_match(method, url)
    assert route is None


@pytest.mark.parametrize("method,pattern,url", MATCHING_ROUTES)
def test_router_add_shortcuts(method, pattern, url):
    router = Router()

    fn = getattr(router, f"add_{method}")

    def home():
        return "Hello, World"

    fn(pattern, home)

    route = router.get_match(method.upper(), url)

    assert route is not None
    assert route.handler is home

    value = route.handler()
    assert value == "Hello, World"

    route = router.get_match(FAKE, url)
    assert route is None


@pytest.mark.parametrize("decorator,pattern,url", MATCHING_ROUTES)
def test_router_decorator(decorator, pattern, url):
    router = Router()

    method = getattr(router, decorator)

    @method(pattern)
    def home():
        return "Hello, World"

    route = router.get_match(decorator.upper(), url)

    assert route is not None
    assert route.handler is home

    value = route.handler()
    assert value == "Hello, World"

    route = router.get_match(FAKE, url)
    assert route is None


def test_router_match_any_by_extension():
    router = Router()

    def a():
        ...

    def b():
        ...

    router.add_get("/a/*.js", a)
    router.add_get("/b/*.css", b)

    m = router.get_match(RouteMethod.GET, b"/a/anything/really")
    assert m is None

    m = router.get_match(RouteMethod.GET, b"/a/anything/really.js")
    assert m is not None
    assert m.handler is a
    assert m.values.get("tail") == "anything/really"

    m = router.get_match(RouteMethod.GET, b"/b/anything/really.css")
    assert m is not None
    assert m.handler is b
    assert m.values.get("tail") == "anything/really"


def test_router_match_any_below():
    router = Router()

    def a():
        ...

    def b():
        ...

    def c():
        ...

    def d():
        ...

    router.add_get("/a/*", a)
    router.add_get("/b/*", b)
    router.add_get("/c/*", c)
    router.add_get("/d/*", d)

    m = router.get_match(RouteMethod.GET, b"/a")
    assert m is not None
    assert m.handler is a
    assert m.values.get("tail") == ""

    m = router.get_match(RouteMethod.GET, b"/a/")
    assert m is not None
    assert m.handler is a
    assert m.values.get("tail") == ""

    m = router.get_match(RouteMethod.GET, b"/a/anything/really")
    assert m is not None
    assert m.handler is a
    assert m.values.get("tail") == "anything/really"

    m = router.get_match(RouteMethod.GET, b"/b/anything/really")
    assert m is not None
    assert m.handler is b
    assert m.values.get("tail") == "anything/really"

    m = router.get_match(RouteMethod.GET, b"/c/anything/really")
    assert m is not None
    assert m.handler is c
    assert m.values.get("tail") == "anything/really"

    m = router.get_match(RouteMethod.GET, b"/d/anything/really")
    assert m is not None
    assert m.handler is d
    assert m.values.get("tail") == "anything/really"

    m = router.get_match(RouteMethod.POST, b"/a/anything/really")
    assert m is None

    m = router.get_match(RouteMethod.POST, b"/b/anything/really")
    assert m is None

    m = router.get_match(RouteMethod.POST, b"/c/anything/really")
    assert m is None

    m = router.get_match(RouteMethod.POST, b"/d/anything/really")
    assert m is None


def test_router_match_among_many():
    router = Router()

    def home():
        ...

    def home_verbose():
        ...

    def home_options():
        ...

    def home_connect():
        ...

    def get_foo():
        ...

    def create_foo():
        ...

    def patch_foo():
        ...

    def delete_foo():
        ...

    def ws():
        ...

    router.add_trace("/", home_verbose)
    router.add_options("/", home_options)
    router.add_connect("/", home_connect)
    router.add_get("/", home)
    router.add_get("/foo", get_foo)
    router.add_patch("/foo", patch_foo)
    router.add_post("/foo", create_foo)
    router.add_delete("/foo", delete_foo)
    router.add_ws("/ws", ws)

    m = router.get_match(RouteMethod.GET, b"/")
    assert m is not None
    assert m.handler is home

    m = router.get_match(RouteMethod.TRACE, b"/")
    assert m is not None
    assert m.handler is home_verbose

    m = router.get_match(RouteMethod.CONNECT, b"/")
    assert m is not None
    assert m.handler is home_connect

    m = router.get_match(RouteMethod.OPTIONS, b"/")
    assert m is not None
    assert m.handler is home_options

    m = router.get_match(RouteMethod.POST, b"/")
    assert m is None

    m = router.get_match(RouteMethod.GET, b"/foo")
    assert m is not None
    assert m.handler is get_foo

    m = router.get_match(RouteMethod.POST, b"/foo")
    assert m is not None
    assert m.handler is create_foo

    m = router.get_match(RouteMethod.PATCH, b"/foo")
    assert m is not None
    assert m.handler is patch_foo

    m = router.get_match(RouteMethod.DELETE, b"/foo")
    assert m is not None
    assert m.handler is delete_foo

    m = router.get_match(RouteMethod.GET_WS, b"/ws")
    assert m is not None
    assert m.handler is ws


def test_router_match_ws_get_sharing_path():
    router = Router()

    def home():
        ...

    def ws():
        ...

    router.add_get("/", home)
    router.add_ws("/", ws)

    m = router.get_match(RouteMethod.GET, b"/")
    assert m is not None
    assert m.handler is home

    m = router.get_match(RouteMethod.GET_WS, b"/")
    assert m is not None
    assert m.handler is ws


def test_router_match_among_many_decorators():
    router = Router()

    @router.get("/")
    def home():
        ...

    @router.trace("/")
    def home_verbose():
        ...

    @router.options("/")
    def home_options():
        ...

    @router.connect("/")
    def home_connect():
        ...

    @router.get("/foo")
    def get_foo():
        ...

    @router.post("/foo")
    def create_foo():
        ...

    @router.patch("/foo")
    def patch_foo():
        ...

    @router.delete("/foo")
    def delete_foo():
        ...

    m = router.get_match(RouteMethod.GET, b"/")
    assert m is not None
    assert m.handler is home

    m = router.get_match(RouteMethod.TRACE, b"/")
    assert m is not None
    assert m.handler is home_verbose

    m = router.get_match(RouteMethod.CONNECT, b"/")
    assert m is not None
    assert m.handler is home_connect

    m = router.get_match(RouteMethod.OPTIONS, b"/")
    assert m is not None
    assert m.handler is home_options

    m = router.get_match(RouteMethod.POST, b"/")
    assert m is None

    m = router.get_match(RouteMethod.GET, b"/foo")
    assert m is not None
    assert m.handler is get_foo

    m = router.get_match(RouteMethod.POST, b"/foo")
    assert m is not None
    assert m.handler is create_foo

    m = router.get_match(RouteMethod.PATCH, b"/foo")
    assert m is not None
    assert m.handler is patch_foo

    m = router.get_match(RouteMethod.DELETE, b"/foo")
    assert m is not None
    assert m.handler is delete_foo


def test_router_match_with_trailing_slash():
    router = Router()

    def get_foo():
        ...

    def create_foo():
        ...

    router.add_get("/foo", get_foo)
    router.add_post("/foo", create_foo)

    m = router.get_match(RouteMethod.GET, b"/foo/")

    assert m is not None
    assert m.handler is get_foo

    m = router.get_match(RouteMethod.POST, b"/foo/")

    assert m is not None
    assert m.handler is create_foo

    m = router.get_match(RouteMethod.POST, b"/foo//")

    assert m is None


def test_fallback_route():
    router = Router()

    def not_found_handler():
        pass

    router.fallback = not_found_handler
    assert isinstance(router.fallback, Route)
    assert router.fallback.handler is not_found_handler

    m = router.get_match(RouteMethod.POST, b"/")

    assert m is not None
    assert m.handler is not_found_handler


def test_fallback_route_must_be_callable_or_route():
    router = Router()

    def not_found_handler():
        pass

    router.fallback = Route("*", not_found_handler)
    router.fallback = not_found_handler

    class Example:
        def __call__(self):
            pass

    router.fallback = Example()

    with pytest.raises(ValueError):
        router.fallback = False

    with pytest.raises(ValueError):
        router.fallback = "Something"


@pytest.mark.parametrize(
    "first_route,second_route",
    [
        ("/", "/"),
        (b"/", b"/"),
        (b"/", "/"),
        ("/", b"/"),
        ("/home/", "/home"),
        (b"/home/", b"/home"),
        ("/home", "/home/"),
        (b"/home", b"/home/"),
        ("/home", "/home//"),
        (b"/home", b"/home//"),
        ("/hello/world", "/hello/world/"),
        (b"/hello/world", b"/hello/world//"),
        ("/a/b", "/a/b"),
    ],
)
def test_duplicate_pattern_raises(first_route, second_route):
    router = Router()

    def home():
        ...

    def another():
        ...

    router.add_get(first_route, home)

    with pytest.raises(RouteDuplicate):
        router.add_get(second_route, another)


def test_duplicate_pattern_star_raises():
    router = Router()

    def home():
        ...

    def another():
        ...

    router.add_get("*", home)

    with pytest.raises(RouteDuplicate):
        router.add_get("*", another)


def test_more_than_one_star_raises():
    router = Router()

    def home():
        ...

    with pytest.raises(RouteException):
        router.add_get("*/*", home)


def test_automatic_pattern_with_ellipsis():
    router = Router()

    @router.get(...)
    def home():
        ...

    @router.get(...)
    def another():
        ...

    match = router.get_match("GET", "/")
    assert match is None

    match = router.get_match("GET", "/home")

    assert match is not None
    assert match.handler is home

    match = router.get_match("GET", "/another")

    assert match is not None
    assert match.handler is another


def test_automatic_pattern_with_ellipsis_name_normalization():
    router = Router()

    @router.get(...)
    def hello_world():
        ...

    match = router.get_match("GET", "/hello_world")

    assert match is None

    match = router.get_match("GET", "/hello-world")

    assert match is not None
    assert match.handler is hello_world


def test_automatic_pattern_with_ellipsis_index_name():
    router = Router()

    @router.get(...)
    def index():
        ...

    match = router.get_match("GET", "/")

    assert match is not None
    assert match.handler is index


def test_router_iterable():
    router = Router()

    @router.get("/")
    def home():
        ...

    @router.trace("/")
    def home_verbose():
        ...

    @router.options("/")
    def home_options():
        ...

    routes = list(router)
    assert len(routes) == 3

    handlers = {home, home_verbose, home_options}

    for route in routes:
        assert route.handler in handlers

    def fallback():
        ...

    router.fallback = fallback

    routes = list(router)
    assert len(routes) == 4

    handlers = {home, home_verbose, home_options, fallback}

    for route in routes:
        assert route.handler in handlers


@pytest.mark.parametrize(
    "route_pattern,expected_pattern",
    [
        ["/", "/"],
        ["/api/v1/help", "/api/v1/help"],
        ["/api/cats/:cat_id", "/api/cats/{cat_id}"],
        ["/api/cats/:cat_id/friends", "/api/cats/{cat_id}/friends"],
        [
            "/api/cats/:cat_id/friends/:friend_id",
            "/api/cats/{cat_id}/friends/{friend_id}",
        ],
        [
            "/api/cats/{int:cat_id}/friends/{uuid:friend_id}",
            "/api/cats/{cat_id}/friends/{friend_id}",
        ],
        [
            "/api/cats/<int:cat_id>/friends/<uuid:friend_id>",
            "/api/cats/{cat_id}/friends/{friend_id}",
        ],
    ],
)
def test_route_to_openapi_pattern(route_pattern, expected_pattern):
    route = Route(route_pattern, object())

    assert route.mustache_pattern == expected_pattern


@pytest.mark.parametrize(
    "mount_path,expected_scope_path",
    [
        ("", "/*"),
        ("/", "/*"),
        ("/*", "/*"),
        ("/foo", "/foo/*"),
        ("/foo*", "/foo/*"),
        ("/foo/", "/foo/*"),
        ("/foo/*", "/foo/*"),
        ("/admin", "/admin/*"),
        ("/admin*", "/admin/*"),
        ("/a/b/c", "/a/b/c/*"),
        ("/a/b/c/", "/a/b/c/*"),
        ("/a/b/c*", "/a/b/c/*"),
        ("/a/b/c/*", "/a/b/c/*"),
    ],
)
def test_mount_add_method(mount_path, expected_scope_path):
    class ASGIHandler:
        def __call__(self, *args, **kwargs):
            pass

    app = ASGIHandler()
    mount = MountRegistry()
    mount.mount(mount_path, app)

    assert any(
        mount_route.pattern.decode() == expected_scope_path
        for mount_route in mount.mounted_apps
    )
    assert any(mount_route.handler is app for mount_route in mount.mounted_apps)


def test_mount_mounted_paths():
    mount = MountRegistry()
    assert mount.mounted_paths == set()

    mount.mount("/foo", Application())
    assert mount.mounted_paths == {"/foo"}

    mount.mount("/oFo", Application())
    assert mount.mounted_paths == {"/foo", "/ofo"}

    mount.mount("/ooF", Application())
    assert mount.mounted_paths == {"/foo", "/ofo", "/oof"}


def test_mount_add_raise_error_if_path_exist():
    with pytest.raises(AssertionError):
        mount = MountRegistry()
        mount.mount("/foo", None)
        mount.mount("/foo", None)



from asyncio import AbstractEventLoop
from datetime import datetime
from pathlib import Path
from typing import List

import pkg_resources
import pytest
from essentials.folders import get_file_extension

from blacksheep import Request
from blacksheep.common.files.asyncfs import FileContext, FilesHandler
from blacksheep.exceptions import BadRequest, InvalidArgument
from blacksheep.ranges import Range, RangePart
from blacksheep.server.files import (
    FileInfo,
    RangeNotSatisfiable,
    ServeFilesOptions,
    _get_requested_range,
    get_default_extensions,
    get_range_file_getter,
    validate_source_path,
)
from blacksheep.server.files.dynamic import get_response_for_file
from blacksheep.server.files.static import get_response_for_static_content
from blacksheep.server.responses import text
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from blacksheep.utils.aio import get_running_loop


def get_folder_path(folder_name: str) -> str:
    return pkg_resources.resource_filename(__name__, f"./{folder_name}")


def get_file_path(file_name, folder_name: str = "files") -> str:
    return pkg_resources.resource_filename(__name__, f"./{folder_name}/{file_name}")


files2_index_path = get_file_path("index.html", "files2")


@pytest.fixture(scope="module")
def files2_index_contents():
    with open(files2_index_path, mode="rb") as actual_file:
        return actual_file.read()


@pytest.mark.asyncio
async def test_get_response_for_file_raise_for_file_not_found():
    with pytest.raises(FileNotFoundError):
        get_response_for_file(
            FilesHandler(), Request("GET", b"/example.txt", None), "example.txt", 1200
        )


TEST_FILES = [
    get_file_path("lorem-ipsum.txt"),
    get_file_path("example.txt"),
    get_file_path("pexels-photo-126407.jpeg"),
]
TEST_FILES_METHODS = [[i, "GET"] for i in TEST_FILES] + [
    [i, "HEAD"] for i in TEST_FILES
]


@pytest.mark.asyncio
@pytest.mark.parametrize("file_path", TEST_FILES)
async def test_get_response_for_file_returns_file_contents(file_path):
    response = get_response_for_file(
        FilesHandler(), Request("GET", b"/example", None), file_path, 1200
    )

    assert response.status == 200
    data = await response.read()

    with open(file_path, mode="rb") as test_file:
        contents = test_file.read()

    assert data == contents


@pytest.mark.asyncio
@pytest.mark.parametrize("file_path,method", TEST_FILES_METHODS)
async def test_get_response_for_file_returns_headers(file_path, method):
    response = get_response_for_file(
        FilesHandler(), Request(method, b"/example", None), file_path, 1200
    )

    assert response.status == 200

    info = FileInfo.from_path(file_path)
    expected_headers = {
        b"etag": info.etag.encode(),
        b"last-modified": str(info.modified_time).encode(),
        b"accept-ranges": b"bytes",
        b"cache-control": b"max-age=1200",
    }

    for expected_header_name, expected_header_value in expected_headers.items():
        value = response.get_single_header(expected_header_name)

        assert value is not None
        assert value == expected_header_value


@pytest.mark.asyncio
@pytest.mark.parametrize("file_path,method", TEST_FILES_METHODS)
async def test_get_response_for_file_returns_not_modified_handling_if_none_match_header(
    file_path, method
):
    info = FileInfo.from_path(file_path)

    response = get_response_for_file(
        FilesHandler(),
        Request(method, b"/example", [(b"If-None-Match", info.etag.encode())]),
        file_path,
        1200,
    )

    assert response.status == 304
    data = await response.read()
    assert data is None


@pytest.mark.asyncio
@pytest.mark.parametrize("file_path", TEST_FILES)
async def test_get_response_for_file_with_head_method_returns_empty_body_with_info(
    file_path,
):
    response = get_response_for_file(
        FilesHandler(), Request("HEAD", b"/example", None), file_path, 1200
    )

    assert response.status == 200
    data = await response.read()
    assert data is None


@pytest.mark.asyncio
@pytest.mark.parametrize("cache_time", [100, 500, 1200])
async def test_get_response_for_file_returns_cache_control_header(cache_time):
    response = get_response_for_file(
        FilesHandler(), Request("GET", b"/example", None), TEST_FILES[0], cache_time
    )

    assert response.status == 200
    header = response.get_single_header(b"cache-control")

    assert header == f"max-age={cache_time}".encode()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "range_value,expected_bytes,expected_content_range",
    [
        [b"bytes=0-10", b"Lorem ipsu", b"bytes 0-10/447"],
        [b"bytes=10-20", b"m dolor si", b"bytes 10-20/447"],
        [b"bytes=33-44", b"ctetur adip", b"bytes 33-44/447"],
        [b"bytes=15-50", b"or sit amet, consectetur adipiscing", b"bytes 15-50/447"],
        [
            b"bytes=66-",
            b"usmod tempor incididunt ut labore et dolore magna\naliqua. Ut enim ad minim veniam, quis nostrud "
            b"exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis\n aute irure dolor in "
            b"reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint\n "
            b"occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.",
            b"bytes 66-446/447",
        ],
        [
            b"bytes=381-",
            b"nt, sunt in culpa qui officia deserunt mollit anim id est laborum.",
            b"bytes 381-446/447",
        ],
        [
            b"bytes=-50",
            b"a qui officia deserunt mollit anim id est laborum.",
            b"bytes 397-446/447",
        ],
        [
            b"bytes=-66",
            b"nt, sunt in culpa qui officia deserunt mollit anim id est laborum.",
            b"bytes 381-446/447",
        ],
    ],
)
async def test_text_file_range_request_single_part(
    range_value, expected_bytes, expected_content_range
):
    file_path = get_file_path("example.txt")
    response = get_response_for_file(
        FilesHandler(),
        Request("GET", b"/example", [(b"Range", range_value)]),
        file_path,
        1200,
    )
    assert response.status == 206
    body = await response.read()
    assert body == expected_bytes

    assert response.get_single_header(b"content-range") == expected_content_range


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "range_value",
    [
        b"bytes=0-10000000000",
        b"bytes=100-200000",
        b"bytes=1111111111114-",
        b"bytes=-1111111111114",
    ],
)
async def test_invalid_range_request_range_not_satisfiable(range_value):
    file_path = get_file_path("example.txt")
    with pytest.raises(RangeNotSatisfiable):
        get_response_for_file(
            FilesHandler(),
            Request("GET", b"/example", [(b"Range", range_value)]),
            file_path,
            1200,
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "range_value,expected_bytes_lines",
    [
        [
            b"bytes=0-10, 10-20",
            [
                b"--##BOUNDARY##",
                b"Content-Type: text/plain",
                b"Content-Range: bytes 0-10/447",
                b"",
                b"Lorem ipsu",
                b"--##BOUNDARY##",
                b"Content-Type: text/plain",
                b"Content-Range: bytes 10-20/447",
                b"",
                b"m dolor si",
                b"--##BOUNDARY##--",
            ],
        ],
        [
            b"bytes=0-10, -66",
            [
                b"--##BOUNDARY##",
                b"Content-Type: text/plain",
                b"Content-Range: bytes 0-10/447",
                b"",
                b"Lorem ipsu",
                b"--##BOUNDARY##",
                b"Content-Type: text/plain",
                b"Content-Range: bytes 381-446/447",
                b"",
                b"nt, sunt in culpa qui officia deserunt mollit anim id est laborum.",
                b"--##BOUNDARY##--",
            ],
        ],
    ],
)
async def test_text_file_range_request_multi_part(
    range_value: bytes, expected_bytes_lines: List[bytes]
):
    file_path = get_file_path("example.txt")
    response = get_response_for_file(
        FilesHandler(),
        Request("GET", b"/example", [(b"Range", range_value)]),
        file_path,
        1200,
    )
    assert response.status == 206
    content_type = response.content.type
    boundary = content_type.split(b"=")[1]
    body = await response.read()

    expected_bytes_lines = [
        line.replace(b"##BOUNDARY##", boundary) for line in expected_bytes_lines
    ]
    assert body.splitlines() == expected_bytes_lines


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "range_value,matches",
    [
        [b"bytes=0-10", True],
        [b"bytes=0-10", False],
        [b"bytes=10-20", True],
        [b"bytes=10-20", False],
    ],
)
async def test_text_file_range_request_single_part_if_range_handling(
    range_value, matches
):
    file_path = get_file_path("example.txt")
    info = FileInfo.from_path(file_path)

    response = get_response_for_file(
        FilesHandler(),
        Request(
            "GET",
            b"/example",
            [
                (b"Range", range_value),
                (b"If-Range", info.etag.encode() + (b"" if matches else b"xx")),
            ],
        ),
        file_path,
        1200,
    )

    expected_status = 206 if matches else 200

    assert response.status == expected_status

    if not matches:
        body = await response.read()

        with open(file_path, mode="rb") as actual_file:
            assert body == actual_file.read()


@pytest.mark.asyncio
async def test_serve_files_no_discovery(app):
    # Note the folder files3 does not contain an index.html page
    app.serve_files(get_folder_path("files3"))

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.asyncio
async def test_serve_files_deprecated_serve_files_options(files2_index_contents, app):
    with pytest.deprecated_call():
        app.serve_files(ServeFilesOptions(get_folder_path("files2")))  # type: ignore

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert files2_index_contents == await response.read()


@pytest.mark.asyncio
async def test_serve_files_fallback_document(files2_index_contents: bytes, app):
    """Feature used to serve SPAs that use HTML5 History API"""
    app.serve_files(get_folder_path("files2"), fallback_document="index.html")

    await app.start()

    for path in {"/", "/one", "/one/two", "/one/two/anything.txt"}:
        scope = get_example_scope("GET", path, [])
        await app(
            scope,
            MockReceive(),
            MockSend(),
        )

        response = app.response
        assert response.status == 200
        assert await response.read() == files2_index_contents


@pytest.mark.asyncio
async def test_serve_files_serves_index_html_by_default(files2_index_contents, app):
    app.serve_files(get_folder_path("files2"))

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert files2_index_contents == await response.read()


@pytest.mark.asyncio
async def test_serve_files_can_disable_index_html_by_default(app):
    app.serve_files(get_folder_path("files2"), index_document=None)

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.asyncio
async def test_serve_files_custom_index_page(app):
    # Note the folder files3 does not contain an index.html page
    app.serve_files(get_folder_path("files3"), index_document="lorem-ipsum.txt")

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    with open(get_file_path("lorem-ipsum.txt", "files3"), mode="rt") as actual_file:
        content = actual_file.read()
        assert content == await response.text()


@pytest.mark.asyncio
async def test_cannot_serve_files_outside_static_folder(app):
    folder_path = get_folder_path("files")
    app.serve_files(folder_path, discovery=True, extensions={".py"})

    await app.start()

    scope = get_example_scope("GET", "../test_files_serving.py", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.asyncio
async def test_cannot_serve_files_with_unhandled_extension(app):
    folder_path = get_folder_path("files2")
    app.serve_files(folder_path, discovery=True, extensions={".py"})

    await app.start()

    scope = get_example_scope("GET", "/example.config", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.asyncio
async def test_can_serve_files_with_relative_paths(files2_index_contents, app):
    folder_path = get_folder_path("files2")
    app.serve_files(folder_path, discovery=True)

    await app.start()

    scope = get_example_scope("GET", "/styles/../index.html", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.read()
    assert body == files2_index_contents

    scope = get_example_scope("GET", "/styles/fonts/../../index.html", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.read()
    assert body == files2_index_contents

    scope = get_example_scope("GET", "/styles/../does-not-exist.html", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 404


@pytest.mark.parametrize("folder_name", ["files", "files2"])
@pytest.mark.asyncio
async def test_serve_files_discovery(folder_name: str, app):
    folder_path = get_folder_path(folder_name)
    app.serve_files(folder_path, discovery=True)
    extensions = get_default_extensions()

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.text()

    folder = Path(folder_path)
    for item in folder.iterdir():
        if item.is_dir():
            assert f"/{item.name}" in body
            continue

        file_extension = get_file_extension(str(item))
        if file_extension in extensions:
            assert f"/{item.name}" in body
        else:
            assert item.name not in body


@pytest.mark.asyncio
async def test_serve_files_discovery_subfolder(app):
    folder_path = get_folder_path("files2")
    app.serve_files(folder_path, discovery=True)

    await app.start()

    scope = get_example_scope("GET", "/scripts", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.text()

    folder = Path(folder_path) / "scripts"
    for item in folder.iterdir():
        assert f"/{item.name}" in body


@pytest.mark.asyncio
async def test_serve_files_with_custom_files_handler(app):
    file_path = files2_index_path

    with open(file_path, mode="rt") as actual_file:
        expected_body = actual_file.read()

    class CustomFilesHandler(FilesHandler):
        def __init__(self) -> None:
            self.calls = []

        def open(self, file_path: str, mode: str = "rb") -> FileContext:
            self.calls.append(file_path)
            return super().open(file_path, mode)

    app.files_handler = CustomFilesHandler()

    folder_path = get_folder_path("files2")
    app.serve_files(folder_path, discovery=True)

    await app.start()

    scope = get_example_scope("GET", "/index.html", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.text()
    assert body == expected_body

    assert app.files_handler.calls[0] == file_path


def test_file_context_mode_property():
    handler = FilesHandler()
    file_path = files2_index_path
    context = handler.open(file_path, loop=get_running_loop())
    assert context.mode == "rb"
    assert context.loop is not None
    assert isinstance(context.loop, AbstractEventLoop)


@pytest.mark.asyncio
async def test_serve_files_multiple_folders(files2_index_contents, app):
    files1 = get_folder_path("files")
    files2 = get_folder_path("files2")
    app.serve_files(files1, discovery=True, root_path="one")
    app.serve_files(files2, discovery=True, root_path="two")

    await app.start()

    scope = get_example_scope("GET", "/", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 404

    scope = get_example_scope("GET", "/example.txt", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 404

    scope = get_example_scope("GET", "/one/example.txt", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.read()

    with open(get_file_path("example.txt"), mode="rb") as actual_file:
        assert body == actual_file.read()

    scope = get_example_scope("GET", "/two/styles/main.css", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.read()

    with open(get_file_path("styles/main.css", "files2"), mode="rb") as actual_file:
        assert body == actual_file.read()

    scope = get_example_scope("GET", "/two/index.html", [])
    await app(
        scope,
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    body = await response.read()
    assert body == files2_index_contents


def test_validate_source_path_raises_for_invalid_path():
    with pytest.raises(InvalidArgument):
        validate_source_path("./not-existing")

    with pytest.raises(InvalidArgument):
        validate_source_path(files2_index_path)


def test_get_requested_range_raises_for_invalid_range():
    request = Request("GET", b"/foo", [(b"range", b"XXXXXXXXXXXX")])

    with pytest.raises(BadRequest):
        _get_requested_range(request)


@pytest.mark.asyncio
async def test_get_range_file_getter_raises_for_invalid():
    getter = get_range_file_getter(
        FilesHandler(), files2_index_path, 100, Range("bytes", [RangePart(None, None)])
    )

    with pytest.raises(BadRequest):
        async for chunk in getter():
            ...


def test_range_with_unknown_unit_is_ignored():
    request = Request(
        "GET", b"/foo", [(b"Range", b"peanuts=200-1000, 2000-6576, 19000- ")]
    )

    assert _get_requested_range(request) is None


def test_file_info():
    info = FileInfo.from_path(files2_index_path)

    assert info.mime == "text/html"
    data = info.to_dict()

    assert data["size"] == info.size
    assert data["etag"] == info.etag
    assert data["mime"] == info.mime
    assert data["modified_time"] == info.modified_time

    assert info.mime in repr(info)


@pytest.mark.asyncio
async def test_get_response_for_static_content_returns_given_bytes():
    response = get_response_for_static_content(
        Request("GET", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
    )

    assert response.status == 200
    data = await response.read()
    assert data == b"Lorem ipsum dolor sit amet\n"


@pytest.mark.asyncio
async def test_get_response_for_static_content_handles_head():
    response = get_response_for_static_content(
        Request("GET", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
    )

    assert response.status == 200

    head_response = get_response_for_static_content(
        Request("HEAD", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
    )

    data = await head_response.read()
    assert data is None

    for name in {b"content-length", b"content-type"}:
        # Note: a response with content has these headers handled automatically
        assert head_response.get_single_header(name) is not None


@pytest.mark.asyncio
async def test_get_response_for_static_content_can_disable_max_age():
    response = get_response_for_static_content(
        Request("GET", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
        cache_time=-1,
    )

    assert response.status == 200
    assert response.headers.contains(b"Cache-Control") is False

    response = get_response_for_static_content(
        Request("GET", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
        cache_time=20,
    )
    assert response.status == 200
    assert b"20" in response.headers.get_first(b"Cache-Control")


@pytest.mark.asyncio
async def test_get_response_for_static_content_handles_304():
    response = get_response_for_static_content(
        Request("GET", b"/", None),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
    )

    assert response.status == 200

    response = get_response_for_static_content(
        Request("GET", b"/", [(b"If-None-Match", response.get_first_header(b"etag"))]),
        b"text/plain",
        b"Lorem ipsum dolor sit amet\n",
        datetime(2020, 10, 24).timestamp(),
    )

    assert response.status == 304


@pytest.mark.asyncio
async def test_app_fallback_route_static_files(app):
    called = False

    def not_found_handler():
        nonlocal called
        called = True
        return text("Example", 404)

    app.router.fallback = not_found_handler

    app.serve_files(get_folder_path("files3"))

    await app.start()
    await app(
        get_example_scope("GET", "/not-registered", []), MockReceive(), MockSend()
    )

    response = app.response
    response_text = await response.text()
    assert response.status == 404
    assert called is True
    assert response_text == "Example"


@pytest.mark.asyncio
async def test_app_404_handler_static_files_not_found(app):
    called = False

    @app.exception_handler(404)
    async def not_found_handler(*args):
        nonlocal called
        called = True
        return text("Example", 404)

    app.serve_files(get_folder_path("files3"))

    await app.start()
    await app(
        get_example_scope("GET", "/not-registered", []), MockReceive(), MockSend()
    )

    response = app.response
    response_text = await response.text()
    assert response.status == 404
    assert called is True
    assert response_text == "Example"



from dataclasses import dataclass
from datetime import date, datetime
from enum import IntEnum
from typing import Generic, List, Optional, Sequence, TypeVar, Union
from uuid import UUID

import pytest
from openapidocs.common import Format, Serializer
from openapidocs.v3 import Info, Reference, Schema, ValueFormat, ValueType
from pydantic import BaseModel, HttpUrl, validator
from pydantic.generics import GenericModel
from pydantic.types import NegativeFloat, PositiveInt, condecimal, confloat, conint

from blacksheep.server.application import Application
from blacksheep.server.bindings import FromForm
from blacksheep.server.openapi.common import (
    ContentInfo,
    EndpointDocs,
    OpenAPIEndpointException,
    ResponseInfo,
)
from blacksheep.server.openapi.exceptions import (
    DuplicatedContentTypeDocsException,
    UnsupportedUnionTypeException,
)
from blacksheep.server.openapi.v3 import (
    DataClassTypeHandler,
    OpenAPIHandler,
    PydanticModelTypeHandler,
    check_union,
)
from blacksheep.server.routing import RoutesRegistry

T = TypeVar("T")
U = TypeVar("U")


@dataclass
class DateTest:
    one: date
    two: datetime


class PydExampleWithSpecificTypes(BaseModel):
    url: HttpUrl


class PydCat(BaseModel):
    id: int
    name: str


class PydPaginatedSetOfCat(BaseModel):
    items: List[PydCat]
    total: int


class PydTypeWithChildModels(BaseModel):
    child: PydPaginatedSetOfCat
    friend: PydExampleWithSpecificTypes


class PlainClass:
    id: int
    name: str


@dataclass
class Cat:
    id: int
    name: str


@dataclass
class CreateCatInput:
    name: str


@dataclass
class CatOwner:
    id: int
    first_name: str
    last_name: str


@dataclass
class CatDetails(Cat):
    owner: CatOwner
    friends: List[int]


@dataclass
class Combo(Generic[T, U]):
    item_one: T
    item_two: U


@dataclass
class PaginatedSet(Generic[T]):
    items: List[T]
    total: int


@dataclass
class Validated(Generic[T]):
    data: T
    error: str


@dataclass
class SubValidated(Generic[T]):
    sub: Validated[T]


class FooLevel(IntEnum):
    BASIC = 1
    MEDIUM = 2
    SUPER = 3


@dataclass
class Foo:
    a: str
    b: bool
    level: Optional[FooLevel] = None


@dataclass
class CreateFooInput:
    a: str
    b: bool
    level: Optional[FooLevel] = None


@dataclass
class Ufo:
    b: bool
    c: str


@dataclass
class ForwardRefExample:
    value: "PaginatedSet[Cat]"


@dataclass
class GenericWithForwardRefExample(Generic[T]):
    ufo: "Ufo"
    value: "PaginatedSet[T]"


class Error(BaseModel):
    code: int
    message: str


class DataModel(BaseModel):
    numbers: List[int]
    people: List[str]


class PydResponse(GenericModel, Generic[T]):
    data: Optional[T]
    error: Optional[Error]

    @validator("error", always=True)
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v


class PydConstrained(BaseModel):
    a: PositiveInt
    b: NegativeFloat
    big_int: conint(gt=1000, lt=1024)

    big_float: confloat(gt=1000, lt=1024)
    unit_interval: confloat(ge=0, le=1)

    decimal_positive: condecimal(gt=0)
    decimal_negative: condecimal(lt=0)


def get_app() -> Application:
    app = Application()
    app.controllers_router = RoutesRegistry()
    return app


def get_cats_api() -> Application:
    app = Application()
    app.controllers_router = RoutesRegistry()
    get = app.router.get
    post = app.router.post
    delete = app.router.delete

    @get("/api/cats")
    def get_cats() -> PaginatedSet[Cat]:
        ...

    @get("/api/cats/{cat_id}")
    def get_cat_details(cat_id: int) -> CatDetails:
        ...

    @post("/api/cats")
    def create_cat(input: CreateCatInput) -> Cat:
        ...

    @delete("/api/cats/{cat_id}")
    def delete_cat(cat_id: int) -> None:
        ...

    return app


@pytest.fixture(scope="function")
def docs() -> OpenAPIHandler:
    # example documentation
    return OpenAPIHandler(info=Info("Example", "0.0.1"))


class CapitalizeOperationDocs(OpenAPIHandler):
    def get_operation_id(self, docs: Optional[EndpointDocs], handler) -> str:
        return handler.__name__.capitalize().replace("_", " ")


@pytest.fixture
def capitalize_operation_id_docs() -> CapitalizeOperationDocs:
    return CapitalizeOperationDocs(info=Info("Example", "0.0.1"))


@pytest.fixture
def serializer() -> Serializer:
    return Serializer()


@pytest.mark.asyncio
async def test_raises_for_started_app(docs):
    app = Application()

    await app.start()

    with pytest.raises(TypeError):
        docs.bind_app(app)


@pytest.mark.asyncio
async def test_raises_for_duplicated_content_example(docs):
    app = get_app()

    @app.router.get("/")
    @docs(
        responses={
            200: ResponseInfo("Example", content=[ContentInfo(Foo), ContentInfo(Foo)])
        }
    )
    async def example():
        ...

    with pytest.raises(DuplicatedContentTypeDocsException):
        docs.bind_app(app)
        await app.start()


@pytest.mark.asyncio
def test_raises_for_union_type(docs):
    with pytest.raises(UnsupportedUnionTypeException):
        docs.get_schema_by_type(Union[Foo, Ufo])


@pytest.mark.parametrize(
    "annotation,expected_result",
    [
        (Foo, [False, Foo]),
        (Optional[Foo], [True, Foo]),
        (Union[Foo, None], [True, Foo]),
        (Union[None, Foo], [True, Foo]),
    ],
)
def test_check_union(annotation, expected_result):
    assert check_union(annotation) == tuple(expected_result)


def test_register_schema_can_handle_classes_with_same_name(docs):
    @dataclass
    class FooX:
        x: str

    FooX.__name__ = "Foo"

    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(FooX)

    foo_schema = docs.components.schemas["Foo"]
    foox_schema = docs.components.schemas["Foo1"]

    assert foo_schema is not None
    assert foox_schema is not None

    assert isinstance(foo_schema, Schema)
    assert isinstance(foox_schema, Schema)

    assert "x" in foox_schema.properties
    assert "a" in foo_schema.properties


def test_register_schema_handles_repeated_calls(docs):
    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Foo)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 1
    foo_schema = docs.components.schemas["Foo"]
    assert "Foo1" not in docs.components.schemas

    assert foo_schema is not None
    assert isinstance(foo_schema, Schema)
    assert "a" in foo_schema.properties


def test_handles_forward_refs(docs):
    @dataclass
    class Friend:
        foo: "Foo"

    docs.register_schema_for_type(Foo)
    docs.register_schema_for_type(Friend)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 2
    friend_schema = docs.components.schemas["Friend"]

    assert friend_schema is not None
    assert isinstance(friend_schema, Schema)
    assert friend_schema.properties["foo"] == Reference(ref="#/components/schemas/Foo")


def test_register_schema_for_enum(docs):
    docs.register_schema_for_type(FooLevel)

    assert docs.components.schemas is not None
    assert len(docs.components.schemas) == 1
    schema = docs.components.schemas["FooLevel"]

    assert isinstance(schema, Schema)

    assert schema is not None
    assert schema.type == ValueType.INTEGER
    assert schema.enum == [x.value for x in FooLevel]


def test_try_get_schema_for_enum_returns_none_for_not_enum(docs):
    assert docs._try_get_schema_for_enum(Foo) is None


def test_get_parameters_returns_non_for_object_without_binders(docs):
    assert docs.get_parameters(Foo) is None
    assert docs.get_request_body(Foo) is None


def test_get_content_from_response_info_returns_none_for_missing_content(docs):
    assert docs._get_content_from_response_info(None) is None


def test_open_api_handler_object_handlers(docs: OpenAPIHandler):
    assert docs.object_types_handlers is not None
    assert isinstance(docs.object_types_handlers[0], DataClassTypeHandler)
    assert isinstance(docs.object_types_handlers[1], PydanticModelTypeHandler)


def test_open_api_handler_get_fields(docs: OpenAPIHandler):
    assert docs.get_fields(None) == []


@pytest.mark.parametrize(
    "json_path,yaml_path,preferred_format,expected_result",
    [
        ["foo.json", "foo.yaml", Format.YAML, "foo.yaml"],
        ["foo.json", "foo.yaml", Format.JSON, "foo.json"],
        ["/openapi.json", "/openapi.yaml", Format.YAML, "/openapi.yaml"],
    ],
)
def test_get_spec_path_preferred_format(
    json_path, yaml_path, preferred_format, expected_result
):
    docs = OpenAPIHandler(
        info=Info("Example", "0.0.1"),
        json_spec_path=json_path,
        yaml_spec_path=yaml_path,
        preferred_format=preferred_format,
    )

    assert docs.get_spec_path() == expected_result


def test_get_spec_path_raises_for_unsupported_preferred_format(docs):
    docs.preferred_format = "NOPE"  # type: ignore

    with pytest.raises(OpenAPIEndpointException):
        docs.get_spec_path()


def test_dates_handling(docs: OpenAPIHandler, serializer: Serializer):
    docs.register_schema_for_type(DateTest)

    assert docs.components.schemas is not None
    schema = docs.components.schemas["DateTest"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(Application()))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        DateTest:
            type: object
            required:
            - one
            - two
            properties:
                one:
                    type: string
                    format: date
                    nullable: false
                two:
                    type: string
                    format: date-time
                    nullable: false
""".strip()
    )


def test_register_schema_for_generic_with_list(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(PaginatedSet[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["PaginatedSetOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(Application()))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        PaginatedSetOfFoo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Foo'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


def test_register_schema_for_multiple_generic_with_list(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(PaginatedSet[Foo])
    docs.register_schema_for_type(PaginatedSet[Ufo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["PaginatedSetOfFoo"]
    assert isinstance(schema, Schema)

    schema = docs.components.schemas["PaginatedSetOfUfo"]
    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(Application()))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        PaginatedSetOfFoo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Foo'
                total:
                    type: integer
                    format: int64
                    nullable: false
        Ufo:
            type: object
            required:
            - b
            - c
            properties:
                b:
                    type: boolean
                    nullable: false
                c:
                    type: string
                    nullable: false
        PaginatedSetOfUfo:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Ufo'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


def test_register_schema_for_generic_with_property(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(Validated[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["ValidatedOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(Application()))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ValidatedOfFoo:
            type: object
            required:
            - data
            - error
            properties:
                data:
                    $ref: '#/components/schemas/Foo'
                error:
                    type: string
                    nullable: false
""".strip()
    )


def test_register_schema_for_generic_sub_property(
    docs: OpenAPIHandler, serializer: Serializer
):
    docs.register_schema_for_type(Validated[Foo])
    docs.register_schema_for_type(SubValidated[Foo])

    assert docs.components.schemas is not None
    schema = docs.components.schemas["SubValidatedOfFoo"]

    assert isinstance(schema, Schema)

    yaml = serializer.to_yaml(docs.generate_documentation(Application()))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths: {}
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ValidatedOfFoo:
            type: object
            required:
            - data
            - error
            properties:
                data:
                    $ref: '#/components/schemas/Foo'
                error:
                    type: string
                    nullable: false
        SubValidatedOfFoo:
            type: object
            required:
            - sub
            properties:
                sub:
                    $ref: '#/components/schemas/ValidatedOfFoo'
""".strip()
    )


@pytest.mark.asyncio
async def test_register_schema_for_multi_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/combo")
    def combo_example() -> Combo[Cat, Foo]:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /combo:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ComboOfCatAndFoo'
            operationId: combo_example
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        ComboOfCatAndFoo:
            type: object
            required:
            - item_one
            - item_two
            properties:
                item_one:
                    $ref: '#/components/schemas/Cat'
                item_two:
                    $ref: '#/components/schemas/Foo'
""".strip()
    )


@pytest.mark.asyncio
async def test_register_schema_for_generic_with_list_reusing_ref(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/one")
    def one() -> PaginatedSet[Cat]:
        ...

    @app.route("/two")
    def two() -> PaginatedSet[Cat]:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /one:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: one
    /two:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: two
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


def test_get_type_name_raises_for_invalid_object_type(docs: OpenAPIHandler):
    with pytest.raises(ValueError):
        docs.get_type_name(10)


@pytest.mark.asyncio
async def test_handling_of_forward_references(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/")
    def forward_ref_example() -> ForwardRefExample:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/ForwardRefExample'
            operationId: forward_ref_example
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        ForwardRefExample:
            type: object
            required:
            - value
            properties:
                value:
                    $ref: '#/components/schemas/PaginatedSetOfCat'
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_normal_class(docs: OpenAPIHandler, serializer: Serializer):
    """
    Plain classes are simply ignored, since their handling would be ambiguous:
    should the library inspect type annotations, __dict__?
    (in fact, the built-in json module throws for them).
    """
    app = get_app()

    @app.route("/")
    def plain_class() -> PlainClass:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                nullable: false
            operationId: plain_class
components: {}
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_pydantic_class_with_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/")
    def home() -> PydPaginatedSetOfCat:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydPaginatedSetOfCat'
            operationId: home
components:
    schemas:
        PydCat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PydPaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_pydantic_class_with_child_models(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/")
    def home() -> PydTypeWithChildModels:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydTypeWithChildModels'
            operationId: home
components:
    schemas:
        PydCat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PydPaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        PydExampleWithSpecificTypes:
            type: object
            required:
            - url
            properties:
                url:
                    type: string
                    format: uri
                    maxLength: 2083
                    minLength: 1
                    nullable: false
        PydTypeWithChildModels:
            type: object
            required:
            - child
            - friend
            properties:
                child:
                    $ref: '#/components/schemas/PydPaginatedSetOfCat'
                friend:
                    $ref: '#/components/schemas/PydExampleWithSpecificTypes'
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_pydantic_class_in_generic(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/")
    def home() -> PaginatedSet[PydCat]:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfPydCat'
            operationId: home
components:
    schemas:
        PydCat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfPydCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/PydCat'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_sequence(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.route("/")
    def home() -> Sequence[Cat]:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                type: array
                                nullable: false
                                items:
                                    $ref: '#/components/schemas/Cat'
            operationId: home
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
""".strip()
    )


def test_handling_of_generic_with_forward_references(docs: OpenAPIHandler):
    with pytest.warns(UserWarning):
        docs.register_schema_for_type(GenericWithForwardRefExample[Cat])


@pytest.mark.asyncio
async def test_cats_api(docs: OpenAPIHandler, serializer: Serializer):
    app = get_cats_api()
    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: get_cats
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: create_cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: get_cat_details
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: delete_cat
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_cats_api_capital_operations_ids(
    capitalize_operation_id_docs: CapitalizeOperationDocs,
    serializer: Serializer,
):
    app = get_cats_api()
    docs = capitalize_operation_id_docs

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /api/cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: Get cats
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: Create cat
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /api/cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/CatDetails'
            operationId: Get cat details
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
        delete:
            responses:
                '204':
                    description: Success response
            operationId: Delete cat
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
        CreateCatInput:
            type: object
            required:
            - name
            properties:
                name:
                    type: string
                    nullable: false
        CatOwner:
            type: object
            required:
            - id
            - first_name
            - last_name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                first_name:
                    type: string
                    nullable: false
                last_name:
                    type: string
                    nullable: false
        CatDetails:
            type: object
            required:
            - id
            - name
            - owner
            - friends
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
                owner:
                    $ref: '#/components/schemas/CatOwner'
                friends:
                    type: array
                    nullable: false
                    items:
                        type: integer
                        format: int64
                        nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_handling_of_pydantic_types(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.route("/")
    def home() -> PydExampleWithSpecificTypes:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydExampleWithSpecificTypes'
            operationId: home
components:
    schemas:
        PydExampleWithSpecificTypes:
            type: object
            required:
            - url
            properties:
                url:
                    type: string
                    format: uri
                    maxLength: 2083
                    minLength: 1
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_pydantic_generic(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.route("/")
    def home() -> PydResponse[PydCat]:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydResponse[PydCat]'
            operationId: home
components:
    schemas:
        PydCat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        Error:
            type: object
            required:
            - code
            - message
            properties:
                code:
                    type: integer
                    format: int64
                    nullable: false
                message:
                    type: string
                    nullable: false
        PydResponse[PydCat]:
            type: object
            required:
            - data
            - error
            properties:
                data:
                    $ref: '#/components/schemas/PydCat'
                error:
                    $ref: '#/components/schemas/Error'
""".strip()
    )


@pytest.mark.asyncio
async def test_pydantic_constrained_types(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.route("/")
    def home() -> PydConstrained:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PydConstrained'
            operationId: home
components:
    schemas:
        PydConstrained:
            type: object
            required:
            - a
            - b
            - big_int
            - big_float
            - unit_interval
            - decimal_positive
            - decimal_negative
            properties:
                a:
                    type: integer
                    format: int64
                    minimum: 0
                    nullable: false
                b:
                    type: number
                    format: float
                    maximum: 0
                    nullable: false
                big_int:
                    type: integer
                    format: int64
                    maximum: 1024
                    minimum: 1000
                    nullable: false
                big_float:
                    type: number
                    format: float
                    maximum: 1024
                    minimum: 1000
                    nullable: false
                unit_interval:
                    type: number
                    format: float
                    nullable: false
                decimal_positive:
                    type: number
                    format: float
                    minimum: 0
                    nullable: false
                decimal_negative:
                    type: number
                    format: float
                    maximum: 0
                    nullable: false
""".strip()
    )


def test_pydantic_model_handler_does_not_raise_for_array_without_field_info():
    handler = PydanticModelTypeHandler()
    assert handler._open_api_v2_field_schema_to_type(None, {"type": "array"}) is list


def test_pydantic_model_handler_does_not_raise_for_file_type():
    handler = PydanticModelTypeHandler()
    assert handler._open_api_v2_field_schema_to_type(None, {"type": "file"}) == Schema(
        type=ValueType.STRING, format=ValueFormat.BINARY
    )


def test_pydantic_model_handler_defaults_to_empty_schema():
    handler = PydanticModelTypeHandler()
    assert (
        handler._open_api_v2_field_schema_to_type(None, {"type": "unknown"}) == Schema()
    )


def test_pydantic_model_handler_handles_type_without__fields__():
    handler = PydanticModelTypeHandler()

    class Foo:
        @staticmethod
        def schema():
            return {"properties": {"foo": {"type": "boolean"}}}

    handler.get_type_fields(Foo)


@pytest.mark.asyncio
async def test_schema_registration(docs: OpenAPIHandler, serializer: Serializer):
    @docs.register(
        Schema(
            type=ValueType.OBJECT,
            required=["foo"],
            properties={
                "foo": Schema(
                    ValueType.INTEGER, minimum=10, maximum=100, nullable=False
                ),
                "ufo": Schema(
                    ValueType.STRING, min_length=5, max_length=10, nullable=False
                ),
            },
        )
    )
    class A:
        ...

    app = get_app()

    @app.route("/")
    def home() -> A:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/A'
            operationId: home
components:
    schemas:
        A:
            type: object
            required:
            - foo
            properties:
                foo:
                    type: integer
                    maximum: 100
                    minimum: 10
                    nullable: false
                ufo:
                    type: string
                    maxLength: 10
                    minLength: 5
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_handles_ref_for_optional_type(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.route("/cats")
    def one() -> PaginatedSet[Cat]:
        ...

    @app.route("/cats/{cat_id}")
    def two(cat_id: int) -> Optional[Cat]:
        ...

    @app.route("/cats_alt/{cat_id}")
    def three(cat_id: int) -> Cat:
        ...

    @app.route("/cats_value_pattern/{uuid:cat_id}")
    def four(cat_id: UUID) -> Cat:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /cats:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/PaginatedSetOfCat'
            operationId: one
    /cats/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
                '404':
                    description: Object not found
            operationId: two
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /cats_alt/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: three
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: integer
                    format: int64
                    nullable: false
                description: ''
                required: true
    /cats_value_pattern/{cat_id}:
        get:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Cat'
            operationId: four
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: string
                    format: uuid
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        Cat:
            type: object
            required:
            - id
            - name
            properties:
                id:
                    type: integer
                    format: int64
                    nullable: false
                name:
                    type: string
                    nullable: false
        PaginatedSetOfCat:
            type: object
            required:
            - items
            - total
            properties:
                items:
                    type: array
                    nullable: false
                    items:
                        $ref: '#/components/schemas/Cat'
                total:
                    type: integer
                    format: int64
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_handles_from_form_docs(docs: OpenAPIHandler, serializer: Serializer):
    app = get_app()

    @app.router.post("/foo")
    def one(data: FromForm[CreateFooInput]) -> Foo:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /foo:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Foo'
            operationId: one
            parameters: []
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                required: true
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        CreateFooInput:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
""".strip()
    )


@pytest.mark.asyncio
async def test_websockets_routes_are_ignored(
    docs: OpenAPIHandler, serializer: Serializer
):
    app = get_app()

    @app.router.post("/foo")
    def one(data: FromForm[CreateFooInput]) -> Foo:
        ...

    @app.router.ws("/ws")
    def websocket_route() -> None:
        ...

    docs.bind_app(app)
    await app.start()

    yaml = serializer.to_yaml(docs.generate_documentation(app))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Example
    version: 0.0.1
paths:
    /foo:
        post:
            responses:
                '200':
                    description: Success response
                    content:
                        application/json:
                            schema:
                                $ref: '#/components/schemas/Foo'
            operationId: one
            parameters: []
            requestBody:
                content:
                    multipart/form-data:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                    application/x-www-form-urlencoded:
                        schema:
                            $ref: '#/components/schemas/CreateFooInput'
                required: true
components:
    schemas:
        Foo:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
        CreateFooInput:
            type: object
            required:
            - a
            - b
            properties:
                a:
                    type: string
                    nullable: false
                b:
                    type: boolean
                    nullable: false
                level:
                    type: integer
                    nullable: true
                    enum:
                    - 1
                    - 2
                    - 3
""".strip()
    )


@pytest.mark.asyncio
async def test_mount_oad_generation(serializer: Serializer):
    """
    Tests support for OAD generation of mounted apps, using the options:

    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True
    """
    parent = Application(show_error_details=True)
    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True

    docs = OpenAPIHandler(info=Info(title="Parent API", version="0.0.1"))
    docs.bind_app(parent)

    @dataclass
    class CreateCatInput:
        name: str
        email: str
        foo: int

    @dataclass
    class CreateDogInput:
        name: str
        email: str
        example: int

    @dataclass
    class CreateParrotInput:
        name: str
        email: str

    @parent.router.get("/")
    def a_home():
        """Parent root."""
        return "Hello, from the parent app - for information, navigate to /docs"

    @parent.router.get("/cats")
    def get_cats_conflicting():
        """Conflict! This will be overridden by the child app route!"""
        return "CONFLICT"

    child_1 = Application()

    @child_1.router.get("/")
    def get_cats():
        """Gets a list of cats."""
        return "Gets a list of cats."

    @child_1.router.post("/")
    def create_cat(data: CreateCatInput):
        """Creates a new cat."""
        return "Creates a new cat."

    @child_1.router.delete("/{cat_id}")
    def delete_cat(cat_id: str):
        """Deletes a cat by id."""
        return "Deletes a cat by id."

    child_2 = Application()

    @child_2.router.get("/")
    def get_dogs():
        """Gets a list of dogs."""
        return "Gets a list of dogs."

    @child_2.router.post("/")
    def create_dog(data: CreateDogInput):
        """Creates a new dog."""
        return "Creates a new dog."

    @child_2.router.delete("/{dog_id}")
    def delete_dog(dog_id: str):
        """Deletes a dog by id."""
        return "Deletes a dog by id."

    child_3 = Application()

    @child_3.router.get("/")
    def get_parrots():
        """Gets a list of parrots."""
        return "Gets a list of parrots"

    @child_3.router.post("/")
    def create_parrot(data: CreateParrotInput):
        """Creates a new parrot."""
        return "Creates a new parrot"

    @child_3.router.delete("/{parrot_id}")
    def delete_parrot(parrot_id: str):
        """Deletes a parrot by id."""
        return "Deletes a parrot by id."

    parent.mount("/cats", child_1)
    parent.mount("/dogs", child_2)
    parent.mount("/parrots", child_3)

    await parent.start()

    yaml = serializer.to_yaml(docs.generate_documentation(parent))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Parent API
    version: 0.0.1
paths:
    /:
        get:
            responses: {}
            operationId: a_home
            summary: Parent root.
            description: Parent root.
    /cats:
        get:
            responses: {}
            operationId: get_cats
            summary: Gets a list of cats.
            description: Gets a list of cats.
        post:
            responses: {}
            operationId: create_cat
            summary: Creates a new cat.
            description: Creates a new cat.
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateCatInput'
                required: true
    /cats/{cat_id}:
        delete:
            responses: {}
            operationId: delete_cat
            summary: Deletes a cat by id.
            description: Deletes a cat by id.
            parameters:
            -   name: cat_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
    /dogs:
        get:
            responses: {}
            operationId: get_dogs
            summary: Gets a list of dogs.
            description: Gets a list of dogs.
        post:
            responses: {}
            operationId: create_dog
            summary: Creates a new dog.
            description: Creates a new dog.
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateDogInput'
                required: true
    /dogs/{dog_id}:
        delete:
            responses: {}
            operationId: delete_dog
            summary: Deletes a dog by id.
            description: Deletes a dog by id.
            parameters:
            -   name: dog_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
    /parrots:
        get:
            responses: {}
            operationId: get_parrots
            summary: Gets a list of parrots.
            description: Gets a list of parrots.
        post:
            responses: {}
            operationId: create_parrot
            summary: Creates a new parrot.
            description: Creates a new parrot.
            parameters: []
            requestBody:
                content:
                    application/json:
                        schema:
                            $ref: '#/components/schemas/CreateParrotInput'
                required: true
    /parrots/{parrot_id}:
        delete:
            responses: {}
            operationId: delete_parrot
            summary: Deletes a parrot by id.
            description: Deletes a parrot by id.
            parameters:
            -   name: parrot_id
                in: path
                schema:
                    type: string
                    nullable: false
                description: ''
                required: true
components:
    schemas:
        CreateCatInput:
            type: object
            required:
            - name
            - email
            - foo
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
                foo:
                    type: integer
                    format: int64
                    nullable: false
        CreateDogInput:
            type: object
            required:
            - name
            - email
            - example
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
                example:
                    type: integer
                    format: int64
                    nullable: false
        CreateParrotInput:
            type: object
            required:
            - name
            - email
            properties:
                name:
                    type: string
                    nullable: false
                email:
                    type: string
                    nullable: false
""".strip()
    )


@pytest.mark.asyncio
async def test_mount_oad_generation_sub_children(serializer: Serializer):
    """
    Tests support for OAD generation of mounted apps, using the options:

    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True
    """
    parent = Application(show_error_details=True)
    parent.mount_registry.auto_events = True
    parent.mount_registry.handle_docs = True

    docs = OpenAPIHandler(info=Info(title="Parent API", version="0.0.1"))
    docs.bind_app(parent)

    @parent.router.get("/")
    def a_home():
        """Parent root."""
        return "Hello, from the parent app - for information, navigate to /docs"

    child_1 = Application()
    child_2 = Application()
    child_3 = Application()

    @child_1.router.get("/")
    def child_1_home():
        return "Child 1 home."

    @child_2.router.get("/")
    def child_2_home():
        return "Child 2 home."

    @child_3.router.get("/")
    def child_3_home():
        return "Child 3 home."

    child_1.mount_registry.auto_events = True
    child_1.mount_registry.handle_docs = True
    child_2.mount_registry.auto_events = True
    child_2.mount_registry.handle_docs = True

    parent.mount("/child-1", child_1)
    child_1.mount("/child-2", child_2)
    child_2.mount("/child-3", child_3)

    await parent.start()

    yaml = serializer.to_yaml(docs.generate_documentation(parent))

    assert (
        yaml.strip()
        == """
openapi: 3.0.3
info:
    title: Parent API
    version: 0.0.1
paths:
    /:
        get:
            responses: {}
            operationId: a_home
            summary: Parent root.
            description: Parent root.
    /child-1:
        get:
            responses: {}
            operationId: child_1_home
    /child-1/child-2:
        get:
            responses: {}
            operationId: child_2_home
    /child-1/child-2/child-3:
        get:
            responses: {}
            operationId: child_3_home
components: {}
""".strip()
    )



import ntpath
from enum import Enum
from functools import lru_cache
from io import BytesIO
from typing import Any, AnyStr, AsyncIterable, Callable, Union

from blacksheep import Content, JSONContent, Response, StreamedContent, TextContent
from blacksheep.common.files.asyncfs import FilesHandler
from blacksheep.plugins import json as json_plugin

MessageType = Any


class ContentDispositionType(Enum):
    """
    Represents the type for a Content-Disposition header (inline or attachment).
    The Content-Disposition response header is a header indicating if the content
    is expected to be displayed inline in the browser as part of a Web page
    or as an attachment, that is downloaded and saved locally.
    """

    INLINE = "inline"
    ATTACHMENT = "attachment"


def _ensure_bytes(value: AnyStr) -> bytes:
    if isinstance(value, str):
        return value.encode()
    if isinstance(value, bytes):
        return value
    raise ValueError("Input value must be bytes or str")


def _json_serialize(obj) -> str:
    return json_plugin.dumps(obj)


def _json_content(obj) -> JSONContent:
    return JSONContent(obj, _json_serialize)


def _optional_content(message: Any = None) -> Content:
    if isinstance(message, str):
        return TextContent(message)
    else:
        return _json_content(message)


def status_code(status: int = 200, message: Any = None) -> Response:
    """
    Returns a plain response with given status, with optional message;
    sent as plain text or JSON.
    """
    if not message:
        return Response(status)
    return Response(status, content=_optional_content(message))


def ok(message: Any = None) -> Response:
    """
    Returns an HTTP 200 OK response, with optional message;
    sent as plain text or JSON."""
    return status_code(200, message)


def created(message: Any = None, location: AnyStr = "") -> Response:
    """
    Returns an HTTP 201 Created response, to the given location
    and with optional JSON content.
    """
    return Response(
        201,
        [(b"Location", _ensure_bytes(location))] if location else [],
        content=_optional_content(message) if message else None,
    )


def accepted(message: Any = None) -> Response:
    """
    Returns an HTTP 202 Accepted response, with optional message;
    sent as plain text or JSON.
    """
    return status_code(202, message)


def no_content() -> Response:
    """
    Returns an HTTP 204 No Content response.
    """
    return Response(204)


def not_modified() -> Response:
    """
    Returns an HTTP 304 Not Modified response.
    """
    return Response(304)


def unauthorized(message: Any = None) -> Response:
    """
    Returns an HTTP 401 Unauthorized response, with optional message;
    sent as plain text or JSON.
    """
    return status_code(401, message)


def forbidden(message: Any = None) -> Response:
    """
    Returns an HTTP 403 Forbidden response, with optional message;
    sent as plain text or JSON.
    """
    return status_code(403, message)


def bad_request(message: Any = None) -> Response:
    """
    Returns an HTTP 400 Bad Request response, with optional message;
    sent as plain text or JSON.
    """
    return status_code(400, message)


def not_found(message: Any = None) -> Response:
    """
    Returns an HTTP 404 Not Found response, with optional message;
    sent as plain text or JSON.
    """
    return status_code(404, message)


def moved_permanently(location: AnyStr) -> Response:
    """
    Returns an HTTP 301 Moved Permanently response, to the given location.
    """
    return Response(301, [(b"Location", _ensure_bytes(location))])


def redirect(location: AnyStr) -> Response:
    """
    Returns an HTTP 302 Found response (commonly called redirect),
    to the given location.
    """
    return Response(302, [(b"Location", _ensure_bytes(location))])


def see_other(location: AnyStr) -> Response:
    """
    Returns an HTTP 303 See Other response, to the given location.
    """
    return Response(303, [(b"Location", _ensure_bytes(location))])


def temporary_redirect(location: AnyStr) -> Response:
    """
    Returns an HTTP 307 Temporary Redirect response, to the given location.
    """
    return Response(307, [(b"Location", _ensure_bytes(location))])


def permanent_redirect(location: AnyStr) -> Response:
    """
    Returns an HTTP 308 Permanent Redirect response, to the given location.
    """
    return Response(308, [(b"Location", _ensure_bytes(location))])


def text(value: str, status: int = 200) -> Response:
    """
    Returns a response with text/plain content,
    and given status (default HTTP 200 OK).
    """
    return Response(
        status, None, Content(b"text/plain; charset=utf-8", value.encode("utf8"))
    )


def html(value: str, status: int = 200) -> Response:
    """
    Returns a response with text/html content,
    and given status (default HTTP 200 OK).
    """
    return Response(
        status, None, Content(b"text/html; charset=utf-8", value.encode("utf8"))
    )


def json(data: Any, status: int = 200) -> Response:
    """
    Returns a response with application/json content,
    and given status (default HTTP 200 OK).
    """
    return Response(
        status,
        None,
        Content(
            b"application/json",
            json_plugin.dumps(data).encode("utf8"),
        ),
    )


def pretty_json(
    data: Any,
    status: int = 200,
    indent: int = 4,
) -> Response:
    """
    Returns a response with indented application/json content,
    and given status (default HTTP 200 OK).
    """
    return Response(
        status,
        None,
        Content(
            b"application/json",
            json_plugin.pretty_dumps(data).encode("utf8"),
        ),
    )


FileInput = Union[Callable[[], AsyncIterable[bytes]], str, bytes, bytearray, BytesIO]


@lru_cache(2000)
def _get_file_provider(file_path: str) -> Callable[[], AsyncIterable[bytes]]:
    async def data_provider():
        async for chunk in FilesHandler().chunks(file_path):
            yield chunk

    return data_provider


def _file(
    value: FileInput,
    content_type: str,
    content_disposition_type: ContentDispositionType,
    file_name: str = None,
) -> Response:
    if file_name:
        exact_file_name = ntpath.basename(file_name)
        if not exact_file_name:
            raise ValueError(
                "Invalid file name: it should be an exact "
                'file name without path, for example: "foo.txt"'
            )

        content_disposition_value = (
            f'{content_disposition_type.value}; filename="{exact_file_name}"'
        )
    else:
        content_disposition_value = content_disposition_type.value

    content: Content
    content_type_value = _ensure_bytes(content_type)

    if isinstance(value, str):
        # value is treated as a path
        content = StreamedContent(content_type_value, _get_file_provider(value))
    elif isinstance(value, BytesIO):

        async def data_provider():
            try:
                value.seek(0)

                while True:
                    chunk = value.read(1024 * 64)

                    if not chunk:
                        break

                    yield chunk
                yield b""
            finally:
                if not value.closed:
                    value.close()

        content = StreamedContent(content_type_value, data_provider)
    elif callable(value):
        # value is treated as an async generator
        async def data_provider():
            async for chunk in value():
                yield chunk
            yield b""

        content = StreamedContent(content_type_value, data_provider)
    elif isinstance(value, bytes):
        content = Content(content_type_value, value)
    elif isinstance(value, bytearray):
        content = Content(content_type_value, bytes(value))
    else:
        raise ValueError(
            "Invalid value, expected one of: Callable, str, "
            "bytes, bytearray, io.BytesIO"
        )

    return Response(
        200, [(b"Content-Disposition", content_disposition_value.encode())], content
    )


def file(
    value: FileInput,
    content_type: str,
    *,
    file_name: str = None,
    content_disposition: ContentDispositionType = ContentDispositionType.ATTACHMENT,
) -> Response:
    """
    Returns a binary file response with given content type and optional
    file name, for download (attachment)
    (default HTTP 200 OK). This method supports both call with bytes,
    or a generator yielding chunks.

    Remarks: this method does not handle cache, ETag and HTTP 304 Not Modified
    responses; when handling files it is recommended to handle cache, ETag and
    Not Modified, according to use case.
    """
    return _file(value, content_type, content_disposition, file_name)
